import numpy as np
from cython import boundscheck, wraparound
from cpython.array cimport array, clone


@boundscheck(False)
@wraparound(False)
cdef double trace(double[:, :] m, int n) nogil:
    cdef:
        double res = 0.0
        int i
    for i in range(n):
        res += m[i, i]
    return res


@boundscheck(False)
@wraparound(False)
cdef double vectors_dot_prod(double[:] x, double[:] y, int n) nogil:
    cdef:
        double res = 0.0
        int i
    for i in range(n):
        res += x[i] * y[i]
    return res


@boundscheck(False)
@wraparound(False)
cdef double[:] matrix_vector_mult(double[:, :] mat, double[:] vec, int rows, int cols):
    cdef:
        int i
        array[double] result, template = array('d')
    result = clone(template, rows, zero=False)
    for i in range(rows):
        result[i] = vectors_dot_prod(mat[i], vec, cols)
    return result


@boundscheck(False)
cdef double _2x2_det(double[:, :] m) nogil:
    return m[0, 0] * m[1, 1] - m[0, 1] * m[1, 0]


@boundscheck(False)
cdef double _3x3_det(double[:, :] m) nogil:
    return m[0, 0] * (m[1, 1] * m[2, 2] - m[1, 2] * m[2, 1]) \
         - m[0, 1] * (m[1, 0] * m[2, 2] - m[1, 2] * m[2, 0]) \
         + m[0, 2] * (m[1, 0] * m[2, 1] - m[1, 1] * m[2, 0])


@boundscheck(False)
cdef double[:, :] _3x3_inv(double[:, :] m):
    cdef:
        double det = _3x3_det(m)
        double[:, :] inv = np.empty((3, 3), dtype=np.double)
    inv[0, 0] = (m[1, 1] * m[2, 2] - m[2, 1] * m[1, 2]) / det
    inv[0, 1] = (m[2, 1] * m[0, 2] - m[0, 1] * m[2, 2]) / det
    inv[0, 2] = (m[0, 1] * m[1, 2] - m[1, 1] * m[0, 2]) / det
    inv[1, 0] = (m[2, 0] * m[1, 2] - m[1, 0] * m[2, 2]) / det
    inv[1, 1] = (m[0, 0] * m[2, 2] - m[2, 0] * m[0, 2]) / det
    inv[1, 2] = (m[1, 0] * m[0, 2] - m[0, 0] * m[1, 2]) / det
    inv[2, 0] = (m[1, 0] * m[2, 1] - m[2, 0] * m[1, 1]) / det
    inv[2, 1] = (m[2, 0] * m[0, 1] - m[0, 0] * m[2, 1]) / det
    inv[2, 2] = (m[0, 0] * m[1, 1] - m[1, 0] * m[0, 1]) / det
    return inv


@boundscheck(False)
@wraparound(False)
cdef double[:, :] matrix_mult(double[:, :] m1, double[:, :] m2, int rows, int cols, int n):
    cdef:
        int i, j, k
        double[:, :] product = np.empty((rows, cols), dtype=np.double)
    for i in range(rows):
        for j in range(cols):
            product[i][j] = 0.0
            for k in range(n):
                product[i][j] += m1[i][k] * m2[k][j]
    return product


@boundscheck(False)
@wraparound(False)
cdef bint check_orthogonal(double[:, :] m, int rows, int cols, double tol):
    cdef:
        int i = 0, j = 0
        double[:, :] product = matrix_mult(m, m.T, rows, rows, cols)
    for i in range(rows):
        for j in range(rows):
            if i == j:
                if abs(product[i][j] - 1.0) > tol:
                    break
            else:
                if abs(product[i][j]) > tol:
                    break
        if j != rows - 1:
            break
    if i != rows - 1:
        return False
    return True



cdef double trace(double[:, :] m, int n) nogil
cdef double vectors_dot_prod(double[:] x, double[:] y, int n) nogil
cdef double[:] matrix_vector_mult(double[:, :] mat, double[:] vec, int rows, int cols)
cdef double[:, :] matrix_mult(double[:, :] m1, double[:, :] m2, int rows, int cols, int n)
cdef double _2x2_det(double[:, :] m) nogil
cdef double _3x3_det(double[:, :] m) nogil
cdef double[:, :] _3x3_inv(double[:, :] m)
cdef bint check_orthogonal(double[:, :] m, int rows, int cols, double tol)



cdef class Quaternion(object):
    cdef:
        double[:] __quadruple

    cpdef double scalar_part(self)
    cpdef double[:] vector_part(self)
    cdef double[:] __conjugate(self)
    cpdef Quaternion conjugate(self)

    cpdef double _norm(self)
    cpdef double distance(self, Quaternion other)
    cpdef Quaternion versor(self)
    cpdef Quaternion reciprocal(self)

    cpdef double[:, :] real_matrix(self)
    cpdef double complex[:, :] complex_matrix(self)



from ._version import __version__
from .Quaternion import Quaternion
from .UnitQuaternion import UnitQuaternion
from .Rotation import Rotation
from .EulerAnglesConventions import Conventions, Convention
from .EulerAngles import EulerAngles



from BDQuaternions.EulerAnglesConventions cimport Conventions, Convention
from BDQuaternions.EulerAngles cimport EulerAngles
from BDQuaternions.Quaternion cimport Quaternion
from BDQuaternions.UnitQuaternion cimport UnitQuaternion
from BDQuaternions.Rotation cimport Rotation

from BDQuaternions.utils cimport random_rotation, random_unit_quaternion, random_quaternion



import numpy as np

from cython import boundscheck, wraparound

from cpython.array cimport array, clone
from libc.math cimport sin, cos, atan2, sqrt, M_PI, fabs
from libc.float cimport DBL_MIN
from .EulerAnglesConventions cimport Convention
from ._quaternion_operations cimport quaternion_to_rotation_matrix, quaternion_from_rotation_matrix
from .Rotation cimport Rotation

"""
Euler angles conversion algorithms after Ken Shoemake in Graphics Gems IV (Academic Press, 1994), p. 222

All angles are in radians by default
"""

cdef class EulerAngles(object):

    def __init__(self, double[:] euler_angles, Convention convention):
        self.__euler_angles = self.__reduce_euler_angles(euler_angles)
        self.__convention = convention

    cdef double __reduce_angle(self, double angle, bint center=True, bint half=False) nogil:
        """
        Adjusts rotation angle to be in the range [-2*pi; 2*pi]
        :param angle: angle or array-like of input angle
        :param center: if True (default) adjust angle to be within [-pi; pi]
        :param half: if True (default False) adjust angle to be within [0; pi]
        :return: reduced angle or array of reduced angles
        """
        cdef:
            double reduced_angle
        if angle > 2 * M_PI:
            reduced_angle = angle - 2 * M_PI * (angle // (2 * M_PI))
        elif angle < -2 * M_PI:
            reduced_angle = angle + 2 * M_PI * (fabs(angle) // (2 * M_PI))
        else:
            reduced_angle = angle
        if center:
            if reduced_angle < -M_PI:
                reduced_angle += 2 * M_PI
            elif reduced_angle > M_PI:
                reduced_angle -= 2 * M_PI
        if half:
            if reduced_angle < 0:
                reduced_angle = -reduced_angle
        return reduced_angle

    @boundscheck(False)
    @wraparound(False)
    cdef double[:] __reduce_euler_angles(self, double[:] euler_angles):
        cdef:
            array[double] reduced_angles, template = array('d')
        reduced_angles = clone(template, 3, zero=False)
        reduced_angles[0] = self.__reduce_angle(euler_angles[0], center=True, half=False)
        reduced_angles[1] = self.__reduce_angle(euler_angles[1], center=True, half=False)
        reduced_angles[2] = self.__reduce_angle(euler_angles[2], center=True, half=False)
        return reduced_angles

    @property
    def euler_angles(self):
        return self.__euler_angles

    @euler_angles.setter
    def euler_angles(self, double[:] euler_angles):
        self.__euler_angles = self.__reduce_euler_angles(euler_angles)

    @property
    def convention(self):
        return self.__convention

    def __str__(self):
        cdef:
            int i
            str label
        label = 'Euler angles (' + self.convention.label + ' convention)\n'
        for i in range(3):
            label += self.convention.axes_labels[i] + ': %2.2f\n' % self.euler_angles[i]
        return label[:-1]

    cpdef void to_parent_convention(self):
        """
        Theoretically derived conventions can be nested endless.
        This function will bring the angles to the highest level parent convention.
        """
        if self.__convention.__parent != self.__convention:
            self.__euler_angles = self.__reduce_euler_angles(self.__convention.to_parent(self.__euler_angles))
            self.__convention = self.__convention.__parent

    @boundscheck(False)
    @wraparound(False)
    cpdef double[:, :] rotation_matrix(self):
        """
        Convert Euler angles to rotation matrix
        :return: 3x3 rotation matrix as numpy array of floats
        """
        cdef:
            Convention parent_convention = self.__convention
            int inner_axis, parity, repetition, frame
            int i, j, k
            double ci, si, cj, sj, ck, sk, cc, ss, cs, sc
            double[:, :] m = np.empty((3,3), dtype=np.double)
            double[:] euler_angles
            array[double] template = array('d')
        euler_angles = clone(template, 3, zero=False)
        euler_angles[0] = self.__euler_angles[0]
        euler_angles[1] = self.__euler_angles[1]
        euler_angles[2] = self.__euler_angles[2]
        while parent_convention.__parent != parent_convention:
            euler_angles = parent_convention.to_parent(euler_angles)
            parent_convention = parent_convention.__parent
        inner_axis, parity, repetition, frame = parent_convention.__code
        i = parent_convention.__euler_safe_axis[inner_axis]
        j = parent_convention.__euler_next_axis[i + parity]
        k = parent_convention.__euler_next_axis[i - parity + 1]
        if frame:
            euler_angles[0], euler_angles[2] = euler_angles[2], euler_angles[0]
        if parity:
            euler_angles[0], euler_angles[1], euler_angles[2] = -euler_angles[0], -euler_angles[1], -euler_angles[2]
        ci = cos(euler_angles[0])
        si = sin(euler_angles[0])
        cj = cos(euler_angles[1])
        sj = sin(euler_angles[1])
        ck = cos(euler_angles[2])
        sk = sin(euler_angles[2])
        cc = ci * ck
        ss = si * sk
        cs = ci * sk
        sc = si * ck

        if repetition:
            m[i, i] = cj
            m[i, j] = si * sj
            m[i, k] = ci * sj
            m[j, i] = sj * sk
            m[j, j] = cc - cj * ss
            m[j, k] = -sc - cj * cs
            m[k, i] = -sj * ck
            m[k, j] = cs + cj * sc
            m[k, k] = cj * cc - ss
        else:
            m[i, i] = cj * ck
            m[i, j] = sj * sc - cs
            m[i, k] = sj * cc + ss
            m[j, i] = cj * sk
            m[j, j] = sj * ss + cc
            m[j, k] = sj * cs - sc
            m[k, i] = -sj
            m[k, j] = si * cj
            m[k, k] = ci * cj
        return m

    @boundscheck(False)
    @wraparound(False)
    cpdef void from_rotation_matrix(self, double[:, :] m, Convention convention):
        """
        convert rotation matrix to Euler angles
        :param m: 3x3 rotation matrix
        :param convention: Euler angles convention
        :return: ax, ay, az three Euler angles
        """
        cdef:
            Convention parent_convention = convention
            Convention current_convention = convention
            int inner_axis, parity, repetition, frame
            int i, j, k
            double ci, si, cj, sj, ck, sk, sy, cy, ax, ay, az
            double[:] euler_angles
            array[double] template = array('d')
        euler_angles = clone(template, 3, zero=False)
        while parent_convention.__parent != parent_convention:
            parent_convention = parent_convention.__parent
        inner_axis, parity, repetition, frame = parent_convention.__code
        i = parent_convention.__euler_safe_axis[inner_axis]
        j = parent_convention.__euler_next_axis[i + parity]
        k = parent_convention.__euler_next_axis[i - parity + 1]
        if repetition:
            sy = sqrt(m[i, j] * m[i, j] + m[i, k] * m[i, k])
            if sy > DBL_MIN * 4:
                ax = atan2(m[i, j], m[i, k])
                ay = atan2(sy, m[i, i])
                az = atan2(m[j, i], -m[k, i])
            else:
                ax = atan2(-m[j, k], m[j, j])
                ay = atan2(sy, m[i, i])
                az = 0.0
        else:
            cy = sqrt(m[i, i] * m[i, i] + m[j, i] * m[j, i])
            if cy > DBL_MIN * 4:
                ax = atan2(m[k, j], m[k, k])
                ay = atan2(-m[k, i], cy)
                az = atan2(m[j, i], m[i, i])
            else:
                ax = atan2(-m[j, k], m[j, j])
                ay = atan2(-m[k, i], cy)
                az = 0.0
        if parity:
            ax, ay, az = -ax, -ay, -az
        if frame:
            ax, az = az, ax
        euler_angles[0] = ax
        euler_angles[1] = ay
        euler_angles[2] = az
        while parent_convention != convention:
            while current_convention.__parent != parent_convention:
                current_convention = current_convention.__parent
            euler_angles = current_convention.from_parent(euler_angles)
            parent_convention = current_convention
            current_convention = convention
        self.__euler_angles = euler_angles
        self.__convention = current_convention


    cpdef void change_convention(self, Convention new_convention):
        """
        Convert Euler angles from given to new convention
        :param new_convention: dict describing new Euler angles convention
        :return: ax, ay, az three Euler angles
        """
        cdef:
            double[:, :] m
        m = self.rotation_matrix()
        self.from_rotation_matrix(m, new_convention)


    cdef double[:] __to_quaternion(self):
        """
        Convert Euler angles to quaternion
        :return: Quaternion as an array of for floats [w*1, x*i, y*j, z*k]
        """
        cdef:
            double[:, :] matrix
        matrix = self.rotation_matrix()
        return quaternion_from_rotation_matrix(matrix)

    cpdef Rotation to_quaternion(self):
        return Rotation(self.__to_quaternion(), euler_angles_convention=self.__convention)


    cdef void __from_quaternion(self, double[:] quadruple, Convention convention):
        """
        Convert Quaternion to Euler angles
        :param quadruple: Quaternion as an array of for floats [w*1, x*i, y*j, z*k]
        :param convention: dict describing Euler angles convention
        """
        cdef:
            double[:, :] matrix
        matrix = quaternion_to_rotation_matrix(quadruple)
        self.from_rotation_matrix(matrix, convention)


    cpdef void from_quaternion(self, Rotation quaternion, Convention convention):
        self.__from_quaternion(quaternion.quadruple, convention)



import numbers
import numpy as np

from cython import boundscheck, wraparound

from .Quaternion cimport Quaternion
from ._quaternion_operations cimport exp as q_exp, log as q_log
from libc.math cimport exp as c_exp, log as c_log

@boundscheck(False)
@wraparound(False)
cpdef exp(arg):
    """
    Calculate exponent function on quaternions and numbers
    :param arg: Quaternion, number or array of both or mix
    :return: exponent of Quaternion, number or array of both or mix
    """
    cdef:
        int i, size
    if isinstance(arg, Quaternion):
        return Quaternion(q_exp(arg.quadruple))
    elif isinstance(arg, numbers.Number):
        return c_exp(arg)
    elif isinstance(arg, (list, tuple, np.ndarray)):
        result = np.copy(np.array(arg))
        shape = result.shape
        size = result.size
        result = result.ravel()
        for i in range(size):
            result[i] = exp(result[i])
        return result.reshape(shape)
    else:
        raise ValueError('Not supported argument of type %s' % str(type(arg)))

@boundscheck(False)
@wraparound(False)
cpdef log(arg):
    """
    Calculate logarithm function on quaternions and numbers
    :param arg: Quaternion, number or array of both or mix
    :return: logarithm of Quaternion, number or array of both or mix
    """
    cdef:
        int i, size
    if isinstance(arg, Quaternion):
        return Quaternion(q_log(arg.quadruple))
    elif isinstance(arg, numbers.Number):
        return c_log(arg)
    elif isinstance(arg, (list, tuple, np.ndarray)):
        result = np.copy(np.array(arg))
        shape = result.shape
        size = result.size
        result = result.ravel()
        for i in range(size):
            result[i] = log(result[i])
        return result.reshape(shape)
    else:
        raise ValueError('Not supported argument of type %s' % str(type(arg)))



from .UnitQuaternion cimport UnitQuaternion
from .EulerAnglesConventions cimport Convention


cdef class Rotation(UnitQuaternion):
    cdef Convention __euler_angles_convention

    cpdef Rotation conjugate(self)
    cpdef Rotation reciprocal(self)
    cpdef double[:] rotate_vector(self, double[:] xyz)
    cpdef double[:, :] rotate(self, double[:, :] xyz)



import numpy as np

from cython import boundscheck, wraparound

from cpython.array cimport array, clone
from cpython.object cimport Py_EQ, Py_NE
from libc.math cimport M_PI


"""
description of rotations after Ken Shoemake in Graphics Gems IV (Academic Press, 1994), p. 222
"""

cdef class Conventions(object):

    def __init__(self):
        # the tuples are for inner axis (X - 0, Y - 1, Z - 2), parity (Even - 0, Odd - 1),
        # repetition (No - 0, Yes - 1), frame (0 - static; 1 - rotating frame)
        self.__euler_angles_codes = {
            # static frame
            'XYZs': (0, 0, 0, 0), 'XYXs': (0, 0, 1, 0), 'XZYs': (0, 1, 0, 0),
            'XZXs': (0, 1, 1, 0), 'YZXs': (1, 0, 0, 0), 'YZYs': (1, 0, 1, 0),
            'YXZs': (1, 1, 0, 0), 'YXYs': (1, 1, 1, 0), 'ZXYs': (2, 0, 0, 0),
            'ZXZs': (2, 0, 1, 0), 'ZYXs': (2, 1, 0, 0), 'ZYZs': (2, 1, 1, 0),
            # rotating frame
            'XYZr': (0, 0, 0, 1), 'XYXr': (0, 0, 1, 1), 'XZYr': (0, 1, 0, 1),
            'XZXr': (0, 1, 1, 1), 'YZXr': (1, 0, 0, 1), 'YZYr': (1, 0, 1, 1),
            'YXZr': (1, 1, 0, 1), 'YXYr': (1, 1, 1, 1), 'ZXYr': (2, 0, 0, 1),
            'ZXZr': (2, 0, 1, 1), 'ZYXr': (2, 1, 0, 1), 'ZYZr': (2, 1, 1, 1)
        }

        self.__default_convention = 'XYZs'

        """
        Dictionary of Euler angles conventions
        variants should be lower-cased possible list of synonyms of the convention
        """
        self.__special_conventions = {
            # special named conventions
            'Nautical': {'variants': ('nautical', 'aircraft', 'cardan'),
                         'axes': 'ZYXr',
                         'axes_labels': ('roll', 'pitch', 'yaw'),
                         'labels': ('yaw', 'pitch', 'roll'),
                         'description': 'Nautical (yaw pitch roll) ZYXr convention'},
            'Bunge': {'variants': ('bunge',),
                      'axes': 'ZXZr',
                      'axes_labels': ('X', 'Y', 'Z'),
                      'labels': ('phi1', 'Phi', 'phi2'),
                      'description': 'Bunge (phi1 Phi phi2) ZXZr convention'},
            'Matthies': {'variants': ('matthies', 'nfft', 'abg'),
                         'axes': 'ZYZr',
                         'axes_labels': ('X', 'Y', 'Z'),
                         'labels': ('alpha', 'beta', 'gamma'),
                         'description': 'Matthies (alpha beta gamma) ZYZr convention'},
            'Roe': {'variants': ('roe',),
                    'axes': 'ZYZr',
                    'axes_labels': ('RD', 'TD', 'ND'),
                    'labels': ('Psi', 'Theta', 'Phi'),
                    'description': 'Roe (Psi, Theta, Phi) RD,TD,ND convention (ZYZr)'},
        }

        self.__general_conventions = {
            # General conventions in the axes names notation XYZr or ZXZs. r/s means rotating or static frame
            # static frame conventions
            'XYZs': {'variants': ('xyzs', 'sxyz'),
                     'axes': 'XYZs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XYZ static frame convention'},
            'XYXs': {'variants': ('xyxs', 'sxyx'),
                     'axes': 'XYXs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XYX static frame convention'},
            'XZYs': {'variants': ('xzys', 'sxzy'),
                     'axes': 'XZYs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XZY static frame convention'},
            'XZXs': {'variants': ('xzxs', 'sxzx'),
                     'axes': 'XZXs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XZX static frame convention'},
            'YZXs': {'variants': ('yzxs', 'syzx'),
                     'axes': 'YZXs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YZX static frame convention'},
            'YZYs': {'variants': ('yzys', 'syzy'),
                     'axes': 'YZYs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YZY static frame convention'},
            'YXZs': {'variants': ('yxzs', 'syxz'),
                     'axes': 'YXZs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YXZ static frame convention'},
            'YXYs': {'variants': ('yxys', 'syxy'),
                     'axes': 'YXYs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YXY static frame convention'},
            'ZXYs': {'variants': ('zxys', 'szxy'),
                     'axes': 'ZXYs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZXY static frame convention'},
            'ZXZs': {'variants': ('zxzs', 'szxz'),
                     'axes': 'ZXZs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZXZ static frame convention'},
            'ZYXs': {'variants': ('zyxs', 'szyx'),
                     'axes': 'ZYXs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZYX static frame convention'},
            'ZYZs': {'variants': ('zyzs', 'szyz'),
                     'axes': 'ZYZs',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZYZ static frame convention'},
            # rotating frame
            'ZYXr': {'variants': ('zyxr', 'rzyx'),
                     'axes': 'ZYXr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZYX rotating frame convention'},
            'XYXr': {'variants': ('xyxr', 'rxyx'),
                     'axes': 'XYXr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XYX rotating frame convention'},
            'YZXr': {'variants': ('yzxr', 'ryzx'),
                     'axes': 'YZXr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YZX rotating frame convention'},
            'XZXr': {'variants': ('xzxr', 'rxzx'),
                     'axes': 'XZXr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XZX rotating frame convention'},
            'XZYr': {'variants': ('xzyr', 'rxzy'),
                     'axes': 'XZYr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XZY rotating frame convention'},
            'YZYr': {'variants': ('yzyr', 'ryzy'),
                     'axes': 'YZYr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YZY rotating frame convention'},
            'ZXYr': {'variants': ('zxyr', 'rzxy'),
                     'axes': 'ZXYr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZXY rotating frame convention'},
            'YXYr': {'variants': ('yxyr', 'ryxy'),
                     'axes': 'YXYr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YXY rotating frame convention'},
            'YXZr': {'variants': ('yxzr', 'ryxz'),
                     'axes': 'YXZr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'YXZ rotating frame convention'},
            'ZXZr': {'variants': ('zxzr', 'rzxz'),
                     'axes': 'ZXZr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZXZ rotating frame convention'},
            'XYZr': {'variants': ('xyzr', 'rxyz'),
                     'axes': 'XYZr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'XYZ rotating frame convention'},
            'ZYZr': {'variants': ('zyzr', 'rzyz'),
                     'axes': 'ZYZr',
                     'axes_labels': ('X', 'Y', 'Z'),
                     'labels': ('theta_1', 'theta_2', 'theta_3'),
                     'description': 'ZYZ rotating frame convention'}
        }

        """
        Dictionary of derived Euler angles conventions.
        Those that use rotations with changing directions,
        but can be derived from another 'standard' rotational Euler angles convention.
        Variants should be lower-cased possible list of synonyms of the convention.
        from_parent is the routine to convert from parent convention Euler angles to the described.
        to_parent is the routine to convert to parent convention Euler angles from the described.
        """
        self.__derived_conventions = {
            # 'symmetrical angles' conventions used in crystallographic texture analysis
            'Kocks': {'variants': ('kocks',),
                      'parent_convention': 'Roe',
                      'axes_labels': ('RD', 'TD', 'ND'),
                      'labels': ('Psi', 'Theta', 'phi'),
                      'from_parent': Kocks2Roe(),
                      'to_parent': Kocks2Roe(),
                      'description': 'Kocks (Psi Theta phi) convention'},
            'Canova': {'variants': ('canova',),
                       'parent_convention': 'Roe',
                       'axes_labels': ('RD', 'TD', 'ND'),
                       'labels': ('omega', 'Theta', 'phi'),
                       'from_parent': Canova2Roe(),
                       'to_parent': Canova2Roe(),
                       'description': 'Canova (omega, Theta, phi) convention'}
        }

    @property
    def euler_angles_codes(self):
        return self.__euler_angles_codes

    @property
    def default_convention(self):
        return self.__default_convention

    @property
    def special_conventions(self):
        return self.__special_conventions

    @property
    def general_conventions(self):
        return self.__general_conventions

    @property
    def derived_conventions(self):
        return self.__derived_conventions

    def list_euler_angles_conventions(self, filter=None):
        """
        Function to get list of available conventions
        :param filter: None to select all or list of filtering strings: 'general', 'special', 'derived'
        :return: list of conventions
        """
        general = list(self.general_conventions.keys())
        special = list(self.special_conventions.keys())
        derived = list(self.derived_conventions.keys())
        if filter is None:
            final_list = general + special + derived
        elif isinstance(filter, str):
            if 'general' == filter.strip():
                final_list = general
            elif 'special' == filter.strip():
                final_list = special
            elif 'derived' == filter.strip():
                final_list = derived
            else:
                final_list = general + special + derived
                print('Expected list or any of known filter words [\'general\', \'special\', \'derived\'] or None.')
                print('Falling back to complete list')
        elif isinstance(filter, (list, tuple, np.ndarray)):
            final_list = []
            if 'general' in filter:
                final_list += general
            if 'special' in filter:
                final_list += special
            if 'derived' in filter:
                final_list += derived
            if not final_list:
                final_list = general + special + derived
                print('Expected list or any of known filter words [\'general\', \'special\', \'derived\'] or None.')
                print('Falling back to complete list')
        else:
            final_list = general + special + derived
            print('Expected list of known filter words [\'general\', \'special\', \'derived\'] or None.')
            print('Falling back to complete list')
        return final_list

    @boundscheck(False)
    @wraparound(False)
    cpdef bint check(self, str convention):
        cdef:
            dict conventions
            str key
        conventions = dict(self.__general_conventions, **self.__special_conventions)
        # first we search requested convention in the dict of 'standard' conventions
        for key in conventions.keys():
            if convention.lower().strip() in conventions[key]['variants']:
                return True
        # If not found we look through the dict of 'derived' conventions
        for key in self.__derived_conventions.keys():
            if convention.lower().strip() in self.__derived_conventions[key]['variants']:
                return True
        return False

    @boundscheck(False)
    @wraparound(False)
    cdef dict __get_convention(self, str convention):
        cdef:
            dict conventions, euler_angles_convention, parent_convention
            bint match = False
            str key, parent_label
        conventions = dict(self.__general_conventions, **self.__special_conventions)
        euler_angles_convention = conventions[self.__default_convention]
        # first we search requested convention in the dict of 'standard' conventions
        for key in conventions.keys():
            if convention.lower().strip() in conventions[key]['variants']:
                euler_angles_convention = conventions[key]
                euler_angles_convention['label'] = key
                euler_angles_convention['parent_convention'] = ''
                euler_angles_convention['code'] = self.__euler_angles_codes[euler_angles_convention['axes']]
                euler_angles_convention['to_parent'] = Function()
                euler_angles_convention['from_parent'] = Function()
                match = True
                break
        # If not found we look through the dict of 'derived' conventions
        for key in self.__derived_conventions.keys():
            if convention.lower().strip() in self.__derived_conventions[key]['variants']:
                euler_angles_convention = self.__derived_conventions[key]
                euler_angles_convention['label'] = key
                parent_label = euler_angles_convention['parent_convention']
                while True:
                    parent_convention = self.__get_convention(parent_label)
                    if 'axes' in list(parent_convention.keys()):
                        euler_angles_convention['axes'] = parent_convention['axes']
                        euler_angles_convention['code'] = self.__euler_angles_codes[euler_angles_convention['axes']]
                        break
                match = True
                break
        if not match:
            euler_angles_convention = self.__get_convention(self.__default_convention)
        return euler_angles_convention

    @boundscheck(False)
    @wraparound(False)
    cpdef Convention get_convention(self, str convention):
        """
        returns euler_angles convention as a dict
        :param convention: string with short form of convention e.g. 'XYZs' or 'Bunge'
        :return: dict describing convention
        """
        cdef:
            dict euler_angles_convention
            bint match = False
            str key
        euler_angles_convention = self.__get_convention(convention)
        return Convention(euler_angles_convention['label'], euler_angles_convention['axes'],
                          list(euler_angles_convention['axes_labels']), list(euler_angles_convention['labels']),
                          list(euler_angles_convention['code']), euler_angles_convention['description'],
                          euler_angles_convention['parent_convention'],
                          euler_angles_convention['to_parent'],
                          euler_angles_convention['from_parent'])


cdef class Convention(object):

    def __init__(self, str label, str axes, list axes_labels, list angle_labels, list code, str description='',
                 str parent='', Function to_parent=Function(), Function from_parent=Function()):
        self.__euler_safe_axis = (0, 1, 2, 0)
        self.__euler_next_axis = (1, 2, 0, 1)
        self.__label = label
        self.__axes = axes
        self.__axes_labels = [str(axes_label) for axes_label in axes_labels]
        self.__angle_labels = [str(angle_label) for angle_label in angle_labels]
        self.__code = [int(code_i) for code_i in code]
        self.__description = description
        self.__conventions = Conventions()
        self.__to_parent = to_parent
        self.__from_parent = from_parent
        if self.__conventions.check(parent):
            self.__parent = self.__conventions.get_convention(parent)
        else:
            self.__parent = self

    @property
    def euler_next_axis(self):
        return self.__euler_next_axis

    @property
    def label(self):
        return self.__label

    @property
    def axes(self):
        return self.__axes

    @property
    def axes_labels(self):
        return self.__axes_labels

    @property
    def angle_labels(self):
        return self.__angle_labels

    @property
    def code(self):
        return self.__code

    @property
    def description(self):
        return self.__description

    @property
    def parent(self):
        return self.__parent

    def __richcmp__(x, y, int op):
        if op == Py_EQ:
            if isinstance(x, Convention) and isinstance(y, Convention):
                if x.label == y.label:
                    if x.code == y.code:
                        if x.axes == y.axes:
                            if x.axes_labels == y.axes_labels:
                                if x.angle_labels == y.angle_labels:
                                    return True
            return False
        elif op == Py_NE:
            if isinstance(x, Convention) and isinstance(y, Convention):
                if x.label == y.label:
                    if x.code == y.code:
                        if x.axes == y.axes:
                            if x.axes_labels == y.axes_labels:
                                if x.angle_labels == y.angle_labels:
                                    return False
            return True
        else:
            return False

    @property
    def parent(self):
        if self.__parent != self:
            return self.__parent
        else:
            return None

    cpdef double[:] to_parent(self, double[:] euler_angles):
        return self.__to_parent.evaluate(euler_angles)

    cpdef double[:] from_parent(self, double[:] euler_angles):
        return self.__from_parent.evaluate(euler_angles)

    @boundscheck(False)
    @wraparound(False)
    cpdef void print_convention_tree(self):
        """
        Theoretically derived conventions can be nested endless.
        This function prints out the tree beginning from highest level parent convention.
        """
        cdef:
            list flat_parent_list = []
            Convention current_convention = self
        while current_convention.parent:
            flat_parent_list.append(current_convention)
            current_convention = current_convention.parent
        flat_parent_list.append(current_convention)
        level = 0
        while flat_parent_list:
            current_convention = flat_parent_list.pop()
            print('-' * level + ' ' * (level > 0) + current_convention.label)
            level += 1


cdef class Function(object):

    @boundscheck(False)
    cpdef double[:] evaluate(self, double[:] euler_angles):
        cdef:
            array[double] result, template = array('d')
        result = clone(template, 3, zero=False)
        result[0] = euler_angles[0]
        result[1] = euler_angles[1]
        result[2] = euler_angles[2]
        return result

cdef class Kocks2Roe(Function):

    @boundscheck(False)
    cpdef double[:] evaluate(self, double[:] euler_angles):
        cdef:
            array[double] result, template = array('d')
        result = clone(template, 3, zero=False)
        result[0] = euler_angles[0]
        result[1] = euler_angles[1]
        result[2] = M_PI - euler_angles[2]
        return result

cdef class Canova2Roe(Function):

    @boundscheck(False)
    cpdef double[:] evaluate(self, double[:] euler_angles):
        cdef:
            array[double] result, template = array('d')
        result = clone(template, 3, zero=False)
        result[0] = M_PI / 2 - euler_angles[0]
        result[1] = euler_angles[1]
        result[2] = 3 * M_PI / 2 - euler_angles[2]
        return result



import warnings
import numpy as np

from cython import boundscheck, wraparound

from cpython.array cimport array, clone
from libc.math cimport fabs, sqrt, cos, sin, acos, atan2
from libc.math cimport exp as c_exp, log as c_log
from libc.float cimport DBL_MIN
from scipy.linalg.cython_lapack cimport dsyevd
from ._helpers cimport _3x3_det, trace, check_orthogonal


@wraparound(False)
@boundscheck(False)
cpdef double[:] mul(double[:] q1, double[:] q2):
    """
    Multiplication of two quaternions
    :param q1: first quaternion as any iterable of four numbers
    :param q2: second quaternion as any iterable of four numbers
    :return: result quaternion as numpy array of four floats
    """
    cdef:
        array[double] quadruple, template = array('d')
    quadruple = clone(template, 4, zero=False)
    quadruple[0] = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3]
    quadruple[1] = q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2]
    quadruple[2] = q1[0] * q2[2] - q1[1] * q2[3] + q1[2] * q2[0] + q1[3] * q2[1]
    quadruple[3] = q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1] + q1[3] * q2[0]
    return quadruple


@wraparound(False)
@boundscheck(False)
cpdef double norm(double[:] q) nogil:
    """
    Calculates norm of quaternion
    :param q: quaternion as an iterable of four numbers
    :return: the norm of the quaternion as float number
    """
    return sqrt(q[0] * q[0] + q[1] * q[1] + q[2] * q[2] + q[3] * q[3])


@wraparound(False)
@boundscheck(False)
cpdef double[:, :] real_matrix(double[:] q):
    """
    returns 4x4 real matrix representation of quaternion
        [ a,  b,  c,  d]
        [-b,  a, -d,  c]
        [-c,  d,  a, -b]
        [-d, -c,  b,  a]
    :param q: quaternion as an iterable of four numbers
    :return: 4x4 real matrix as numpy array
    """
    cdef:
        double[:, :] m = np.empty((4, 4), dtype=np.double)
    m[0][0] =  q[0]
    m[0][1] =  q[1]
    m[0][2] =  q[2]
    m[0][3] =  q[3]
    m[1][0] = -q[1]
    m[1][1] =  q[0]
    m[1][2] = -q[3]
    m[1][3] =  q[2]
    m[2][0] = -q[2]
    m[2][1] =  q[3]
    m[2][2] =  q[0]
    m[2][3] = -q[1]
    m[3][0] = -q[3]
    m[3][1] = -q[2]
    m[3][2] =  q[1]
    m[3][3] =  q[0]
    return m


@wraparound(False)
@boundscheck(False)
cpdef double complex[:, :] complex_matrix(double[:] q):
    """
    returns 2x2 complex matrix representation of quaternion
        [ a + b * 1j, c + d * 1j]
        [-c + d * 1j, a - b * 1j]
    :param q: quaternion as an iterable of four numbers
    :return: 2x2 complex matrix as numpy array
    """
    cdef:
        double complex[:, :] cm = np.empty((2, 2), dtype=np.complex128)
    cm[0][0] = q[0] + q[1]* 1j
    cm[0][1] = q[2] + q[3] * 1j
    cm[1][0] = -q[2] + q[3] * 1j
    cm[1][1] = q[0] - q[1] * 1j
    return cm


@wraparound(False)
@boundscheck(False)
cpdef double[:, :] quaternion_to_rotation_matrix(double[:] q):
    """
    Convert versor corresponding to given quaternion to rotation 3x3 matrix
        [1 - 2 * y ** 2 - 2 * z ** 2, 2 * x * y - 2 * w * z, 2 * x * z + 2 * w * y]
        [2 * x * y + 2 * w * z, 1 - 2 * x ** 2 - 2 * z ** 2, 2 * y * z - 2 * w * x]
        [2 * x * z - 2 * w * y, 2 * y * z + 2 * w * x, 1 - 2 * x ** 2 - 2 * y ** 2]
    :param q: quaternion as an iterable of four numbers
    :return: 3x3 rotation matrix as numpy array
    """
    cdef:
        double n = norm(q)
        double w = q[0] / n, x = q[1] / n, y = q[2] / n, z = q[3] / n
        double[:, :] m = np.empty((3, 3), dtype=np.double)
    m[0, 0] = 1 - 2 * y ** 2 - 2 * z ** 2
    m[0, 1] = 2 * x * y - 2 * w * z
    m[0, 2] = 2 * x * z + 2 * w * y
    m[1, 0] = 2 * x * y + 2 * w * z
    m[1, 1] = 1 - 2 * x ** 2 - 2 * z ** 2
    m[1, 2] = 2 * y * z - 2 * w * x
    m[2, 0] = 2 * x * z - 2 * w * y
    m[2, 1] = 2 * y * z + 2 * w * x
    m[2, 2] = 1 - 2 * x ** 2 - 2 * y ** 2
    return m


@wraparound(False)
@boundscheck(False)
cpdef double[:] quaternion_from_rotation_matrix(double[:, :] m):
    """
    Convert 3x3 rotation matrix to quaternion
    :param m: 3x3 rotation matrix as numpy array
    :return: quaternion as numpy array of four floats
    """
    cdef:
        double det_m, t, r, w, s, x, y, z
        array[double] quadruple, template = array('d')
        int n = 4, lwork = 57, liwork = 23, info
        double k_m[4][4]
        double work[57]
        int iwork[23]
        double w_n[4]
        char L = b'L', J = b'V'
    quadruple = clone(template, 4, zero=False)
    det_m = _3x3_det(m)
    if abs(1 - det_m ** 2) > 1e-6:
        raise ValueError('Not a rotation matrix. det M = %2.2g' % det_m)
    if check_orthogonal(m, 3, 3, 1.0e-12) and abs(det_m - 1.0) < 1.0e-12:
        t = trace(m, 3)
        if t > 3 * DBL_MIN:
            r = sqrt(1 + t)
            w = 0.5 * r
            s = 0.5 / r
            x = (m[2, 1] - m[1, 2]) * s
            y = (m[0, 2] - m[2, 0]) * s
            z = (m[1, 0] - m[0, 1]) * s
        elif m[0, 0] >= m[1, 1] and m[0, 0] >= m[2, 2]:
            r = sqrt(1 + m[0, 0] - m[1, 1] - m[2, 2])
            s = 0.5 / r
            w = (m[2, 1] - m[1, 2]) * s
            x = 0.5 * r
            y = (m[0, 1] + m[1, 0]) * s
            z = (m[2, 0] + m[0, 2]) * s
        elif m[1, 1] >= m[2, 2]:
            r = sqrt(1 + m[1, 1] - m[0, 0] - m[2, 2])
            s = 0.5 / r
            w = (m[0, 2] - m[2, 0]) * s
            x = (m[0, 1] + m[1, 0]) * s
            y = 0.5 * r
            z = (m[1, 2] + m[2, 1]) * s
        else:
            r = sqrt(1 + m[2, 2] - m[0, 0] - m[1, 1])
            s = 0.5 / r
            w = (m[1, 0] - m[0, 1]) * s
            x = (m[2, 0] + m[0, 2]) * s
            y = (m[1, 2] + m[2, 1]) * s
            z = 0.5 * r
        quadruple[0] = w
        quadruple[1] = x
        quadruple[2] = y
        quadruple[3] = z
    else:
        warnings.warn('Not a rotation matrix. det M = %2.2g' % det_m)
        k_m[0][0] = (m[0, 0] - m[1, 1] - m[2, 2]) / 3.0
        k_m[0][1] = (m[0, 1] + m[1, 0]) / 3.0
        k_m[0][2] = (m[0, 2] + m[2, 0]) / 3.0
        k_m[0][3] = (m[2, 1] - m[1, 2]) / 3.0
        k_m[1][0] = (m[0, 1] + m[1, 0]) / 3.0
        k_m[1][1] = (m[1, 1] - m[0, 0] - m[2, 2]) / 3.0
        k_m[1][2] = (m[1, 2] + m[2, 1]) / 3.0
        k_m[1][3] = (m[0, 2] - m[2, 0]) / 3.0
        k_m[2][0] = (m[0, 2] + m[2, 0]) / 3.0
        k_m[2][1] = (m[1, 2] + m[2, 1]) / 3.0
        k_m[2][2] = (m[2, 2] - m[0, 0] - m[1, 1]) / 3.0
        k_m[2][3] = (m[1, 0] - m[0, 1]) / 3.0
        k_m[3][0] = (m[2, 1] - m[1, 2]) / 3.0
        k_m[3][1] = (m[0, 2] - m[2, 0]) / 3.0
        k_m[3][2] = (m[1, 0] - m[0, 1]) / 3.0
        k_m[3][3] = (m[0, 0] + m[1, 1] + m[2, 2]) / 3.0
        dsyevd(&J, &L, &n, &k_m[0][0], &n, &w_n[0], &work[0], &lwork, &iwork[0], &liwork, &info)
        quadruple[0] = k_m[3][3]
        quadruple[1] = k_m[3][0]
        quadruple[2] = k_m[3][1]
        quadruple[3] = k_m[3][2]
    return quadruple


@wraparound(False)
@boundscheck(False)
cpdef double[:] exp(double[:] q):
    """
    Calculates exp() function on quaternion
    :param q: quaternion as an iterable of four numbers
    :return: result quaternion as numpy array of four floats
    """
    cdef:
        double v_norm, a1, sin_v_norm, a = q[0], b = q[1], c = q[2], d = q[3]
        array[double] result, template = array('d')
    result = clone(template, 4, zero=False)
    v_norm = sqrt(b * b + c * c + d * d)
    a1 = c_exp(a)
    if v_norm > 0.0:
        sin_v_norm = a1 * sin(v_norm) / v_norm
        result[0] = a1 * cos(v_norm)
        result[1] = b * sin_v_norm
        result[2] = c * sin_v_norm
        result[3] = d * sin_v_norm
    else:
        result[0] = a1
        result[1] = 0.0
        result[2] = 0.0
        result[3] = 0.0
    return result


@wraparound(False)
@boundscheck(False)
cpdef double[:] log(double[:] q):
    """
    Calculates log() function on quaternion
    :param q: quaternion as any iterable of four numbers
    :return: result quaternion as numpy array of four floats
    """
    cdef:
        double q_norm, v_norm, acos_a_q_norm
        double a = q[0], b = q[1], c = q[2], d = q[3]
        array[double] result, template = array('d')
    result = clone(template, 4, zero=False)
    q_norm = norm(q)
    v_norm = sqrt(b * b + c * c + d * d)
    result[0] = c_log(q_norm)
    if v_norm > 0.0:
        acos_a_q_norm = acos(a / q_norm) / v_norm
        result[1] = b * acos_a_q_norm
        result[2] = c * acos_a_q_norm
        result[3] = d * acos_a_q_norm
    else:
        result[1] = 0.0
        result[2] = 0.0
        result[3] = 0.0
    return result



cpdef double[:] mul(double[:] q1, double[:] q2)
cpdef double norm(double[:] quadruple) nogil

cpdef double[:, :] real_matrix(double[:] q)
cpdef double complex[:, :] complex_matrix(double[:] q)

cpdef double[:, :] quaternion_to_rotation_matrix(double[:] q)
cpdef double[:] quaternion_from_rotation_matrix(double[:, :] m)

cpdef double[:] exp(double[:] q)
cpdef double[:] log(double[:] q)



import numbers
import numpy as np

from cython import wraparound, boundscheck

from cpython.object cimport Py_EQ, Py_NE
from libc.math cimport fabs
from libc.float cimport DBL_MIN

from .Quaternion cimport Quaternion
from ._quaternion_operations cimport mul, quaternion_to_rotation_matrix, quaternion_from_rotation_matrix
from ._helpers cimport matrix_vector_mult
from .UnitQuaternion cimport UnitQuaternion
from .EulerAnglesConventions cimport Conventions, Convention
from .EulerAngles cimport EulerAngles


conventions = Conventions()


cdef class Rotation(UnitQuaternion):
    """
    Rotation is the special class on top of UnitQuaternion dealing with 3D rotations
    Rotation restricts quaternion binary operations to ones allowed for rotations group
    """

    def __init__(self, double[:] quadruple=np.array([1, 0, 0, 0], dtype=np.double),
                 Convention euler_angles_convention=conventions.get_convention('Bunge')):
        self.__euler_angles_convention = euler_angles_convention
        super(Rotation, self).__init__(quadruple)

    def __richcmp__(x, y, int op):
        if op == Py_EQ:
            if isinstance(x, Rotation) and isinstance(y, Rotation):
                return np.allclose(x.quadruple, y.quadruple) or np.allclose(x.quadruple, -y.quadruple)
            return False
        elif op == Py_NE:
            if isinstance(x, Rotation) and isinstance(y, Rotation):
                return not (np.allclose(x.quadruple, y.quadruple) or np.allclose(x.quadruple, -y.quadruple))
            return True
        else:
            return NotImplemented

    cpdef Rotation conjugate(self):
        """
        Calculates conjugate for the Rotation quaternion
        :return: Rotation quaternion which is conjugate of current quaternion
        """
        return Rotation(self.__conjugate(), euler_angles_convention=self.__euler_angles_convention)

    cpdef Rotation reciprocal(self):
        """
        for Unit quaternion reciprocal is equal to conjugate
        :return: Rotation quaternion reciprocal to current quaternion
        """
        return self.conjugate()

    """
    roation matrix representation of Rotation quaternion
    rotation_matrix is a get/set property
    """

    @property
    def rotation_matrix(self):
        return quaternion_to_rotation_matrix(self.__quadruple)

    @rotation_matrix.setter
    def rotation_matrix(self, m):
        self.__quadruple = quaternion_from_rotation_matrix(m)

    """
    axis and angle representation of Rotation quaternion
    axis_angle is a get/set property
    """

    @property
    def axis_angle(self):
        _, axis, theta = self.polar
        return axis, theta * 2

    @axis_angle.setter
    def axis_angle(self, axis_angle_components):
        axis, theta = axis_angle_components
        axis = np.array(axis, dtype=np.float)
        axis_norm = np.sqrt(np.sum(axis * axis))
        if axis_norm > 0:
            axis /= axis_norm
        self.polar = 1, axis, theta / 2

    def __str__(self):
        information = 'Rotation quaternion: ' + str(self.quadruple) + '\n'
        information += 'Euler angles: %s\n' % self.euler_angles_convention.description + '\n'
        information += str(self.euler_angles) + '\n'
        information += 'rotation matrix:\n'
        information += str(np.asarray(self.rotation_matrix)) + '\n'
        information += 'rotation axis, angle:\n'
        information += str(self.axis_angle) + '\n'
        return information

    """
        euler angles convention get/set property
    """

    @property
    def euler_angles_convention(self):
        return self.__euler_angles_convention

    @euler_angles_convention.setter
    def euler_angles_convention(self, euler_angles_convention):
        if isinstance(euler_angles_convention, Convention):
            self.__euler_angles_convention = euler_angles_convention
        else:
            convention = conventions.get_convention(str(euler_angles_convention))
            self.__euler_angles_convention = convention

    """
        Euler angles representation of Rotation quaternion
        euler_angles is a get/set property
    """

    @property
    def euler_angles(self):
        cdef EulerAngles ea
        ea = EulerAngles(np.zeros(3, dtype=np.double), convention=self.__euler_angles_convention)
        ea.from_quaternion(self, self.__euler_angles_convention)
        return ea

    @euler_angles.setter
    def euler_angles(self, EulerAngles ea):
        cdef Rotation r
        r = ea.to_quaternion()
        self.__quadruple = r.__quadruple
        self.__euler_angles_convention = r.__euler_angles_convention

    def __add__(self, other):
        return NotImplemented

    def __sub__(self, other):
        return NotImplemented

    def __mul__(x, y):
        if isinstance(x, Rotation) and isinstance(y, Rotation):
            quadruple = mul(x.quadruple, y.quadruple)
            return Rotation(quadruple, euler_angles_convention=x.euler_angles_convention)
        elif isinstance(x, Rotation) and isinstance(y, UnitQuaternion):
            quadruple = mul(x.quadruple, y.quadruple)
            return Rotation(quadruple, euler_angles_convention=x.euler_angles_convention)
        elif isinstance(x, UnitQuaternion) and isinstance(y, Rotation):
            quadruple = mul(x.quadruple, y.quadruple)
            return Rotation(quadruple, euler_angles_convention=y.euler_angles_convention)
        elif isinstance(x, Rotation) and isinstance(y, Quaternion):
            quadruple = mul(x.quadruple, y.quadruple)
            return Quaternion(quadruple)
        elif isinstance(x, UnitQuaternion) and isinstance(y, Rotation):
            quadruple = mul(x.quadruple, y.quadruple)
            return Quaternion(quadruple)
        elif isinstance(x, Rotation) and isinstance(y, numbers.Number):
            if fabs(float(y) - 1) < 4 * DBL_MIN:
                return Rotation(x.quadruple, euler_angles_convention=x.euler_angles_convention)
            elif fabs(float(y) + 1) < 4 * DBL_MIN:
                return Rotation(-1 * x.quadruple, euler_angles_convention=x.euler_angles_convention)
            else:
                return Quaternion(x.quadruple) * y
        elif isinstance(x, numbers.Number) and isinstance(y, Rotation):
            if fabs(float(x) - 1) < 4 * DBL_MIN:
                return Rotation(y.quadruple, euler_angles_convention=y.euler_angles_convention)
            elif fabs(float(x) + 1) < 4 * DBL_MIN:
                return Rotation(-1 * y.quadruple, euler_angles_convention=y.euler_angles_convention)
            else:
                return Quaternion(y.quadruple) * x
        else:
            return NotImplemented

    @boundscheck(False)
    @wraparound(False)
    cpdef double[:] rotate_vector(self, double[:] xyz):
        """
        Apply rotation to vector
        :param xyz: vector
        :return: rotated vector
        """
        return matrix_vector_mult(quaternion_to_rotation_matrix(self.__quadruple), xyz, 3, 3)

    @boundscheck(False)
    @wraparound(False)
    cpdef double[:, :] rotate(self, double[:, :] xyz):
        """
        Apply rotation to vector or array of vectors
        :param xyz: vector or array of vectors
        :return: rotated vector or array of vectors
        """
        cdef:
            int i, j, k, rows = xyz.shape[0]
            double[:, :] m = quaternion_to_rotation_matrix(self.__quadruple)
            double[:, :] product = np.zeros((rows, 3), dtype=np.double)
        for i in range(rows):
            for j in range(3):
                for k in range(3):
                    product[i][j] += xyz[i][k] * m[j][k]
        return product



import numpy as np

from cython import boundscheck, wraparound

from cpython.array cimport array, clone
from libc.stdlib cimport rand, RAND_MAX
from ._quaternion_operations cimport norm
from .Quaternion cimport Quaternion
from .UnitQuaternion cimport UnitQuaternion
from .Rotation cimport Rotation


cpdef Rotation random_rotation():
    """
    Calculates random rotation quaternion
    :return: random Rotation
    """
    cdef:
        double random_quadruple_norm
        array[double] quadruple, template = array('d')
    quadruple = clone(template, 4, zero=False)
    while True:
        quadruple[0] = rand()
        quadruple[1] = rand()
        quadruple[2] = rand()
        quadruple[3] = rand()
        random_quadruple_norm = norm(quadruple)
        if random_quadruple_norm > 0:
            break
    quadruple[0] /= random_quadruple_norm
    quadruple[1] /= random_quadruple_norm
    quadruple[2] /= random_quadruple_norm
    quadruple[3] /= random_quadruple_norm
    return Rotation(quadruple)


cpdef UnitQuaternion random_unit_quaternion():
    """
    Calculates random unit quaternion
    :return: random UnitQuaternion
    """
    cdef:
        double random_quadruple_norm
        array[double] quadruple, template = array('d')
    quadruple = clone(template, 4, zero=False)
    while True:
        quadruple[0] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[1] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[2] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[3] = (rand() / RAND_MAX - 0.5) * 2
        random_quadruple_norm = norm(quadruple)
        if random_quadruple_norm > 0:
            break
    quadruple[0] /= random_quadruple_norm
    quadruple[1] /= random_quadruple_norm
    quadruple[2] /= random_quadruple_norm
    quadruple[3] /= random_quadruple_norm
    return UnitQuaternion(quadruple)


cpdef Quaternion random_quaternion(double quadruple_norm=1.0):
    """
    Calculates random quaternion
    :return: random Quaternion
    """
    cdef:
        double random_quadruple_norm
        array[double] quadruple, template = array('d')
    quadruple = clone(template, 4, zero=False)
    while True:
        quadruple[0] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[1] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[2] = (rand() / RAND_MAX - 0.5) * 2
        quadruple[3] = (rand() / RAND_MAX - 0.5) * 2
        random_quadruple_norm = norm(quadruple)
        if abs(quadruple_norm) > 0.0:
            if random_quadruple_norm > 0:
                break
        else:
            break
    if random_quadruple_norm > 0:
        quadruple[0] = quadruple[0] / random_quadruple_norm * abs(quadruple_norm)
        quadruple[1] = quadruple[1] / random_quadruple_norm * abs(quadruple_norm)
        quadruple[2] = quadruple[2] / random_quadruple_norm * abs(quadruple_norm)
        quadruple[3] = quadruple[3] / random_quadruple_norm * abs(quadruple_norm)
    return Quaternion(quadruple)


@boundscheck(False)
@wraparound(False)
cpdef random_rotations_array(shape):
    """
    Calculates random rotation quaternions array
    :return: random Rotation array of given shape
    """
    rotations = np.empty(shape, dtype=object)
    size = rotations.size
    rotations = rotations.ravel()
    for i in range(size):
        rotations[i] = random_rotation()
    return rotations.reshape(shape)


@boundscheck(False)
@wraparound(False)
cpdef random_unit_quaternions_array(shape):
    """
    Calculates random unit quaternions array
    :return: random UnitQuaternion array of given shape
    """
    unit_quaternions = np.empty(shape, dtype=object)
    size = unit_quaternions.size
    unit_quaternions = unit_quaternions.ravel()
    for i in range(size):
        unit_quaternions[i] = random_unit_quaternion()
    return unit_quaternions.reshape(shape)


@boundscheck(False)
@wraparound(False)
cpdef random_quaternions_array(shape, double quadruple_norm=1.0):
    """
    Calculates random quaternions array
    :return: random Quaternion array of given shape
    """
    quaternions = np.empty(shape, dtype=object)
    size = quaternions.size
    quaternions = quaternions.ravel()
    for i in range(size):
        quaternions[i] = random_quaternion(quadruple_norm=quadruple_norm)
    return quaternions.reshape(shape)



import numbers
import numpy as np

from libc.float cimport DBL_MIN

from .Quaternion cimport Quaternion
from ._quaternion_operations cimport mul, norm


cdef class UnitQuaternion(Quaternion):
    """
    Sub-class of Quaternion to deal with unit quaternions (of norm == 1)
    """

    def __init__(self, double[:] quadruple=np.array([1, 0, 0, 0], dtype=np.double)):
        assert np.allclose(norm(quadruple), 1.0)
        super(UnitQuaternion, self).__init__(quadruple)

    cpdef UnitQuaternion conjugate(self):
        """
        Calculates conjugate for the Unit Quaternion
        :return: Unit Quaternion which is conjugate of current unit quaternion
        """
        return UnitQuaternion(self.__conjugate())

    cpdef UnitQuaternion reciprocal(self):
        """
        for Unit quaternion reciprocal is equal to conjugate
        :return: Unit quaternion reciprocal to current quaternion
        """
        return self.conjugate()

    def __mul__(x, y):
        if isinstance(x, UnitQuaternion) and isinstance(y, UnitQuaternion):
            return UnitQuaternion(mul(x.quadruple, y.quadruple))
        elif isinstance(x, UnitQuaternion) and isinstance(y, Quaternion):
            return Quaternion(mul(x.quadruple, y.quadruple))
        elif isinstance(x, Quaternion) and isinstance(y, UnitQuaternion):
            return Quaternion(mul(x.quadruple, y.quadruple))
        elif isinstance(x, UnitQuaternion) and isinstance(y, numbers.Number):
            if abs(float(y) - 1) < 4 * DBL_MIN:
                return UnitQuaternion(x.quadruple)
            elif abs(float(y) + 1) < 4 * DBL_MIN:
                return UnitQuaternion(-1 * x.quadruple)
            else:
                return Quaternion(x.quadruple) * y
        elif isinstance(x, numbers.Number) and isinstance(y, UnitQuaternion):
            if abs(float(x) - 1) < 4 * DBL_MIN:
                return UnitQuaternion(y.quadruple)
            elif abs(float(x) + 1) < 4 * DBL_MIN:
                return UnitQuaternion(-1 * y.quadruple)
            else:
                return Quaternion(y.quadruple) * x
        else:
            return NotImplemented



from .EulerAnglesConventions cimport Convention
from .Rotation cimport Rotation


cdef class EulerAngles(object):
    cdef:
        double[:] __euler_angles
        Convention __convention

    cdef double __reduce_angle(self, double angle, bint center=*, bint half=*) nogil
    cdef double[:] __reduce_euler_angles(self, double[:] euler_angles)
    cpdef void to_parent_convention(self)
    cpdef double[:, :] rotation_matrix(self)
    cpdef void from_rotation_matrix(self, double[:, :] m, Convention convention)
    cpdef void change_convention(self, Convention new_convention)

    cdef double[:] __to_quaternion(self)
    cpdef Rotation to_quaternion(self)

    cdef void __from_quaternion(self, double[:] quadruple, Convention convention)
    cpdef void from_quaternion(self, Rotation quaternion, Convention convention)



import numpy as np
import numbers

from cython import boundscheck, wraparound

from libc.math cimport sqrt, cos, sin, acos
from cpython.array cimport array, clone
from cpython.object cimport Py_EQ, Py_NE

from ._quaternion_operations cimport mul, norm, real_matrix, complex_matrix


cdef class Quaternion(object):
    """
    General quaternion object
    """

    def __init__(self, double[:] quadruple=np.array([0, 0, 0, 1], dtype=np.double)):
        self.__quadruple = np.empty(4, dtype=np.double)
        self.__quadruple[0] = quadruple[0]
        self.__quadruple[1] = quadruple[1]
        self.__quadruple[2] = quadruple[2]
        self.__quadruple[3] = quadruple[3]

    def __str__(self):
        return str(np.asarray(self.__quadruple))

    def __repr__(self):
        return str(self)

    def __richcmp__(x, y, int op):
        if op == Py_EQ:
            if isinstance(x, Quaternion) and isinstance(y, Quaternion):
                return np.allclose(x.quadruple, y.quadruple)
            return False
        elif op == Py_NE:
            if isinstance(x, Quaternion) and isinstance(y, Quaternion):
                return not np.allclose(x.quadruple, y.quadruple)
            return True
        else:
            return NotImplemented


    @property
    def quadruple(self):
        return np.asarray(self.__quadruple)

    @quadruple.setter
    def quadruple(self, double[:] quadruple):
        self.__quadruple[0] = quadruple[0]
        self.__quadruple[1] = quadruple[1]
        self.__quadruple[2] = quadruple[2]
        self.__quadruple[3] = quadruple[3]

    @wraparound(False)
    @boundscheck(False)
    cpdef double scalar_part(self):
        """
        Calculates scalar part of the Quaternion
        :return: scalar part of the Quaternion
        """
        return self.__quadruple[0]

    @wraparound(False)
    @boundscheck(False)
    cpdef double[:] vector_part(self):
        """
        Calculates vector part of the Quaternion
        :return: vector part of the Quaternion
        """
        return self.__quadruple[1:4]

    @wraparound(False)
    @boundscheck(False)
    cdef double[:] __conjugate(self):
        cdef:
            double[:] v_part = self.vector_part()
            array[double] quadruple, template = array('d')
        quadruple = clone(template, 4, zero=False)
        quadruple[0] = self.scalar_part()
        quadruple[1] = -v_part[0]
        quadruple[2] = -v_part[1]
        quadruple[3] = -v_part[2]
        return quadruple

    cpdef Quaternion conjugate(self):
        """
        Calculates conjugate for the Quaternion
        :return: Quaternion which is conjugate of current quaternion
        """
        return Quaternion(self.__conjugate())

    def __mul__(x, y):
        if isinstance(x, Quaternion) and isinstance(y, Quaternion):
            return Quaternion(mul(x.quadruple, y.quadruple))
        elif isinstance(x, Quaternion) and isinstance(y, numbers.Number):
            return Quaternion(mul(x.quadruple, np.array([np.double(y), 0, 0, 0], dtype=np.double)))
        elif isinstance(x, numbers.Number) and isinstance(y, Quaternion):
            return Quaternion(mul(np.array([np.double(x), 0, 0, 0], dtype=np.double), y.quadruple))
        else:
            return NotImplemented

    def __add__(x, y):
        if isinstance(x, Quaternion) and isinstance(y, Quaternion):
            return Quaternion(x.quadruple + y.quadruple)
        elif isinstance(x, Quaternion) and isinstance(y, numbers.Number):
            return Quaternion(x.quadruple + np.array([np.double(y), 0, 0, 0]))
        elif isinstance(x, numbers.Number) and isinstance(y, Quaternion):
            return Quaternion(y.quadruple + np.array([np.double(x), 0, 0, 0]))
        else:
            return NotImplemented

    def __sub__(x, y):
        if isinstance(x, Quaternion) and isinstance(y, Quaternion):
            return Quaternion(x.quadruple - y.quadruple)
        elif isinstance(x, Quaternion) and isinstance(y, numbers.Number):
            return Quaternion(x.quadruple - np.array([np.double(y), 0, 0, 0]))
        elif isinstance(x, numbers.Number) and isinstance(y, Quaternion):
            return Quaternion(np.array([np.double(x), 0, 0, 0]) - y.quadruple)
        else:
            return NotImplemented

    cpdef double _norm(self):
        """
        Calculates the norm of the Quaternion
        :return: norm of Quaternion
        """
        return norm(self.__quadruple)

    @property
    def norm(self):
        return self._norm()

    cpdef double distance(self, Quaternion other):
        """
        Calculates distance between two quaternions
        :param other: other Quaternion
        :return: distance to other Quaternion
        """
        return (self - other)._norm()

    cpdef Quaternion versor(self):
        """
        Return versor for current quaternion
        :return: Quaternion which is versor for the given quaternion
        """
        return 1 / self._norm() * self

    cpdef Quaternion reciprocal(self):
        """
        Return quaternion reciprocal to given
        """
        return 1 / (self._norm() ** 2) * self.conjugate()

    def __div__(x, y):
        if isinstance(x, Quaternion) and isinstance(y, Quaternion):
            return NotImplemented
        elif isinstance(x, Quaternion) and isinstance(y, numbers.Number):
            return x * (1.0 / np.double(y))
        elif isinstance(x, numbers.Number) and isinstance(y, Quaternion):
            return y.reciprocal() * np.double(x)
        else:
            return NotImplemented

    def __truediv__(x, y):
        if isinstance(x, Quaternion) and isinstance(y, Quaternion):
            return NotImplemented
        elif isinstance(x, Quaternion) and isinstance(y, numbers.Number):
            return x * (1.0 / np.double(y))
        elif isinstance(x, numbers.Number) and isinstance(y, Quaternion):
            return y.reciprocal() * np.double(x)
        else:
            return NotImplemented


    """
        Property to get/set quaternion using polar notation
    """
    @property
    def polar(self):
        cdef:
            double a, v_norm, theta
            double[:] v
            array[double] n_hat, template = array('d')
        n_hat = clone(template, 3, zero=False)
        if not np.allclose(self._norm(), [0.0]):
            a = self.scalar_part()
            v = self.vector_part()
            v_norm = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
            if v_norm > 0.0:
                n_hat[0] = v[0] / v_norm
                n_hat[1] = v[1] / v_norm
                n_hat[2] = v[2] / v_norm
            else:
                n_hat[0] = 0.0
                n_hat[1] = 0.0
                n_hat[2] = 0.0
            theta = acos(a / self._norm())
            return self._norm(), n_hat, theta
        else:
            n_hat[0] = 0.0
            n_hat[1] = 0.0
            n_hat[2] = 0.0
            return 0.0, n_hat, 0.0

    @polar.setter
    def polar(self, polar_components):
        q_norm, n_hat, theta = polar_components
        n_hat = np.array(n_hat, dtype=np.double)
        assert q_norm >= 0
        a = q_norm * cos(theta)
        v = n_hat * q_norm * sin(theta)
        self.__quadruple = np.hstack((a, v))

    def __pow__(x, power, modulo):
        if isinstance(x, Quaternion) and isinstance(power, numbers.Number):
            q_norm, n_hat, theta = x.polar
            result = Quaternion(np.array([0, 0, 0, 1], dtype=np.double))
            result.polar = (q_norm ** power, n_hat, theta * power)
            return result
        else:
            return NotImplemented

    cpdef double[:, :] real_matrix(self):
        """
        Calculates real 4x4 matrix representation of the quaternion
        :return: 4x4 real numpy array matrix
        """
        return real_matrix(self.__quadruple)

    cpdef double complex[:, :] complex_matrix(self):
        """
        Calculates complex 2x2 matrix representation of the quaternion
        :return: 2x2 complex numpy array matrix
        """
        return complex_matrix(self.__quadruple)



from .Quaternion cimport Quaternion
from .UnitQuaternion cimport UnitQuaternion
from .Rotation cimport Rotation

cpdef Rotation random_rotation()
cpdef UnitQuaternion random_unit_quaternion()
cpdef Quaternion random_quaternion(double quadruple_norm=*)
cpdef random_rotations_array(shape)
cpdef random_unit_quaternions_array(shape)
cpdef random_quaternions_array(shape, double quadruple_norm=*)



cdef class Convention(object):
    cdef:
        int[4] __euler_safe_axis
        int[4] __euler_next_axis
        str __label
        str __axes
        list __axes_labels
        list __angle_labels
        int[4] __code
        str __description
        Convention __parent
        Function __to_parent
        Function __from_parent
        Conventions __conventions

    cpdef double[:] to_parent(self, double[:] euler_angles)
    cpdef double[:] from_parent(self, double[:] euler_angles)
    cpdef void print_convention_tree(self)


cdef class Conventions(object):
    cdef:
        dict __euler_angles_codes
        str __default_convention
        dict __special_conventions
        dict __general_conventions
        dict __derived_conventions

    cpdef bint check(self, str convention)
    cdef dict __get_convention(self, str convention)
    cpdef Convention get_convention(self, str convention)


cdef class Function(object):
    cpdef double[:] evaluate(self, double[:] euler_angles)



import sys
from setuptools import setup, find_packages
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext
from Cython.Build import cythonize

from codecs import open
from os import path, remove
import re


here = path.abspath(path.dirname(__file__))
package_name = 'BDQuaternions'
version_file = path.join(here, package_name, '_version.py')
with open(version_file, 'rt') as f:
    version_file_line = f.read()
version_re = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(version_re, version_file_line, re.M)
if mo:
    version_string = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in %s.' % (version_file,))

readme_file = path.join(here, 'README.md')
with open(readme_file, encoding='utf-8') as f:
    long_description = f.read()

extensions = [
    Extension(
        'BDQuaternions._helpers',
        ['BDQuaternions/_helpers.pyx'],
        depends=['BDQuaternions._helpers.pxd'],
    ),
    Extension(
        'BDQuaternions._quaternion_operations',
        ['BDQuaternions/_quaternion_operations.pyx'],
        depends=['BDQuaternions/_quaternion_operations.pxd'],
    ),
    Extension(
        'BDQuaternions.Quaternion',
        ['BDQuaternions/Quaternion.pyx'],
        depends=['BDQuaternions/Quaternion.pxd'],
    ),
    Extension(
        'BDQuaternions.UnitQuaternion',
        ['BDQuaternions/UnitQuaternion.pyx'],
        depends=['BDQuaternions/UnitQuaternion.pxd'],
    ),
    Extension(
        'BDQuaternions.EulerAnglesConventions',
        ['BDQuaternions/EulerAnglesConventions.pyx'],
        depends=['BDQuaternions/EulerAnglesConventions.pxd'],
    ),
    Extension(
        'BDQuaternions.EulerAngles',
        ['BDQuaternions/EulerAngles.pyx'],
        depends=['BDQuaternions/EulerAngles.pxd'],
    ),
    Extension(
        'BDQuaternions.Rotation',
        ['BDQuaternions/Rotation.pyx'],
        depends=['BDQuaternions/Rotation.pxd'],
    ),
    Extension(
        'BDQuaternions.functions',
        ['BDQuaternions/functions.pyx'],
        depends=['BDQuaternions/functions.pxd'],
    ),
    Extension(
        'BDQuaternions.utils',
        ['BDQuaternions/utils.pyx'],
        depends=['BDQuaternions/utils.pxd'],
    ),
]

copt = {'msvc': [],
        'mingw32': [],
        'unix': []}
lopt = {'mingw32': [],
        'unix': []}


# check whether compiler supports a flag
def has_flag(compiler, flagname):
    import tempfile
    from distutils.errors import CompileError
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            res = compiler.compile([f.name], extra_postargs=[flagname])
            for item in res:
                remove(item)
        except CompileError:
            return False
    return True


# filter flags, returns list of accepted flags
def flag_filter(compiler, flags):
    result = []
    for flag in flags:
        if has_flag(compiler, flag):
            result.append(flag)
    return result


class CustomBuildExt(build_ext):
    def build_extensions(self):
        c = self.compiler.compiler_type
        print('Compiler:', c)
        opts = flag_filter(self.compiler, copt.get(c, []))
        lopts = flag_filter(self.compiler, lopt.get(c, []))
        for e in self.extensions:
            e.extra_compile_args = opts
            e.extra_link_args = lopts
        build_ext.build_extensions(self)


setup(
    name=package_name,
    version=version_string,

    description='Package for manipulations with Quaternions',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/bond-anton/BDQuaternions',

    author='Anton Bondarenko',
    author_email='bond.anton@gmail.com',

    license='Apache Software License',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    keywords='Quaternion 3D rotations',

    packages=find_packages(exclude=['demo', 'tests', 'docs', 'contrib', 'venv']),
    ext_modules=cythonize(extensions, compiler_directives={'language_level': sys.version_info[0]}),
    package_data={'BDQuaternions': ['*.pxd']},
    install_requires=['numpy', 'scipy'],
    cmdclass={'build_ext': CustomBuildExt}
)
